import numpy as np
import cv2 as cv
import time
import sys

# Fix path when importing files from different folder
sys.path.append("./")
from feature import (
    findKeyPoints,
    findMatches,
    filterMatches,
    distanceBetweenMatches,
)
from utils import rescaleFrame

KEYPOINTS_METHOD = "ORB"


def featureMatching(
    videoName, videoFps, videoStartIndex, numOfFramesToSkip, distancePerPx, orientation
):
    # Open capture of video at the set start index
    cap = cv.VideoCapture("./Resources/" + videoName)
    cap.set(cv.CAP_PROP_POS_FRAMES, videoStartIndex)

    # Read first frame from capture
    ret, framePrev = cap.read()

    # Find keypoints and descriptors for the first frame
    kpPrev, desPrev = findKeyPoints(framePrev, KEYPOINTS_METHOD)

    frameIndex = 0

    # Save the frame index of the last estimation to keep track of elapsed  time
    frameIndexLastEstimation = -1

    # Save speed at each estimation for data analysis
    speedData = []

    while ret:
        # Read next frame from capture
        ret, frame = cap.read()

        # Ret will be False if there are no more frames in the video
        if not ret:
            return speedData

        # Check if frame index is in interval of set frames between estimations
        if frameIndex % numOfFramesToSkip == 0:
            # Calculate keypoints and descriptors for the current frame
            kp, des = findKeyPoints(frame, KEYPOINTS_METHOD)

            # Find feature matches between previous and current frame
            matchesBetweenFrames = findMatches(
                frame, framePrev, kp, kpPrev, des, desPrev
            )

            # Calculate the average distance between good matches
            distance = distanceBetweenMatches(matchesBetweenFrames, kp, kpPrev)

            # Convert distance per pixles to distance in meters
            distanceInMeters = distancePerPx * distance

            # Calculate how much time has elapsed since last estimation
            time = (frameIndex - frameIndexLastEstimation) / videoFps

            # Use the estimated distance to calculate speed in Y-axis
            speed = distanceInMeters / time

            # Save speed in data array
            speedData.append(speed * orientation)

            # Update previous states
            kpPrev = kp
            desPrev = des
            framePrev = frame
            frameIndexLastEstimation = frameIndex

        # Update frame index
        frameIndex += 1