import numpy as np
import cv2 as cv

# outpy.avi is a perspective transformed video produced by video_perspective_transformation.py
# cap = cv.VideoCapture('Resources/sim2_realtime.mp4')
# cap = cv.VideoCapture('outpy.avi')

DECIMAL = 4

DISTANCE_REMORA = 1500
DISTANCE_SIM = 1500  # CORRECT in SIM each square is 10cm.
DISTANCE_SIM2 = 2100

# Parameteres for Lucas-Kanade optical flow
lucas_kanade = dict(
    winSize=(15, 15),
    maxLevel=2,
    criteria=(cv.TERM_CRITERIA_EPS | cv.TERM_CRITERIA_COUNT, 10, 0.03),
)
# Parameters for shiTomasi corner detection
features = dict(maxCorners=500, qualityLevel=0.3, minDistance=7, blockSize=7)

# Positioning and speed
current_position = [0, 0]
previous_position = [0, 0]
position = [0, 0]
# speed = [0,0]
speedX = 0
speedY = 0
track_length = 10
detect_interval = 5
tracks = []

frameIndexLastEstimation = -1
# Uses video outpy.avi which is created by video_perspective_transformation.py
# outpy.avi is sim2
# cam = cv.VideoCapture('outpy.avi')
# outpy2.avi is remora_new_edited_0.25x.avi
# cam = cv.VideoCapture('outpy2.avi')
# outpy3.avi is remora_new_edited.avi
# cam = cv.VideoCapture('Resources/remora_new_edited_pt.avi')

# cam = cv.VideoCapture("Resources/remora_transformed.avi")

cam = cv.VideoCapture("Resources/sim_realtime.mp4")
# cam = cv.VideoCapture('Resources/sim_realtime_0.5x.mp4')
# cam = cv.VideoCapture('Resources/sim_realtime_2x.mp4')
# cam = cv.VideoCapture('Resources/fouling2.mp4')

# While using 0.25x speed video of remora_new_edited:
# Counted 21 squares (0.03m each), result showed ca. 0.64m, 21*0.03 = 0.63.
# 0.5 meters took 10 seconds. Which gives 0.05 m/s, actual video would be 4x as fast: 0.2 m/s

# While using real-time remora_new_edited:
#
# 12 seconds gave y-position of 1.7 meters. 12*0.2 = 2.4 meters.

# Note: seems like remora_new is moving too fast because the camera is so close to the net.
# In sim2 it worked because the camera is farther away, which means the net moves slower on the camera.
# The calculated speed would remain the same.

frame_index = 0

while True:
    ret, frame = cam.read()
    gray = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
    visualize = frame.copy()

    if len(tracks) > 0:
        image0, image1 = previous_gray, gray
        p0 = np.float32([tr[-1] for tr in tracks]).reshape(-1, 1, 2)
        p1, st, err = cv.calcOpticalFlowPyrLK(image0, image1, p0, None, **lucas_kanade)
        p0r, st, err = cv.calcOpticalFlowPyrLK(image1, image0, p1, None, **lucas_kanade)
        difference = abs(p0 - p0r).reshape(-1, 2).max(-1)
        good = difference < 1
        new_tracks = []
        for tr, (x, y), good_flag in zip(tracks, p1.reshape(-1, 2), good):
            if not good_flag:
                continue
            tr.append((x, y))
            if len(tr) > track_length:
                del tr[0]
            new_tracks.append(tr)
            cv.circle(visualize, (x, y), 2, (0, 255, 0), -1)

        tracks = new_tracks
        # print(self.tracks)
        cv.polylines(visualize, [np.int32(tr) for tr in tracks], False, (0, 255, 0))

    if frame_index % detect_interval == 0:
        mask = np.zeros_like(gray)

        # Set mask
        # mask[:] = 255
        # Mask for output.avi
        # mask[60:90] = 255
        mask[100:200] = 255
        for x, y in [np.int32(tr[-1]) for tr in tracks]:
            cv.circle(mask, (x, y), 5, 0, -1)
        p = cv.goodFeaturesToTrack(gray, mask=mask, **features)
        if p is not None:
            for x, y in np.float32(p).reshape(-1, 2):
                tracks.append([(x, y)])

    # self.tracks[a][b][c]
    # a is for the main tracking points
    # b is 10 for the trail of the point
    # c is the x or y value of the point

    totalDistanceX = 0
    totalDistanceY = 0
    # Check if frame index is in interval of set frames between estimations
    if frame_index % 10 == 0:

        for a in range(len(tracks)):

            if len(tracks[a]) >= 10:
                # Distance in X-direction
                totalDistanceX = totalDistanceX + tracks[a][9][0] - tracks[a][0][0]

                # Distance in Y-direction
                totalDistanceY = totalDistanceY + tracks[a][9][1] - tracks[a][0][1]

        # Check
        if len(tracks) > 0:
            distanceX = totalDistanceX / len(tracks)
            distanceY = totalDistanceY / len(tracks)

            # Calculate how much time has elapsed since last estimation
            time = (frame_index - frameIndexLastEstimation) / 30
            # print(time)
            # Convert distance per pixles to distance in meters
            distanceInMetersX = distanceX * 0.0003125
            distanceInMetersY = distanceY * 0.0003125

            # Use the estimated distance to calculate speed in Y-axis
            speedX = distanceInMetersX / time
            speedY = distanceInMetersY / time

            # Update frame index
            frameIndexLastEstimation = frame_index

        # Add Y-speed to speedata for later analyzing
        # speedData.append(speedY * -orientation)
        print(speedY)
        # Update previous frame
        previous_gray = gray

    # cv.putText(visualize, f'Position: X {round(position[0]/1,DECIMAL)}(m), Y {round(position[1]/1,DECIMAL)}(m)', (20,400), cv.FONT_HERSHEY_COMPLEX, 0.7, (0,255,255), 2)
    cv.putText(
        visualize,
        f"Speed: X {round(speedX,DECIMAL)}(m/s), Y {round(speedY,DECIMAL)}(m/s)",
        (20, 440),
        cv.FONT_HERSHEY_COMPLEX,
        0.7,
        (0, 255, 255),
        2,
    )
    previous_gray = gray
    frame_index += 1
    cv.imshow("lk_track", visualize)

    ch = 0xFF & cv.waitKey(200)
    if ch == 27:
        break
