import numpy as np
import cv2 as cv
import sys
import time

import multiprocessing as mp
import ctypes as c
from os import system, name

# Fix path when importing files from different folder
sys.path.append("./")
from utils import rescaleFrame
from image import cannyEdgeDetecion, findKeyPoints, findMatches, distanceBetweenMatches
from measurements import calculateSpeed, calculatePosition


# CONSTANTS
KEYPOINTS_METHOD = "ORB"
VIDEO_NAME = "./Resources/fouling.mp4"
# Height, width of the frames
VIDEO_RESOLUTION = (1080, 1920)
FRAME_START_INDEX = 700


def clear():
    # Clear console log in windows
    if name == "nt":
        _ = system("cls")


def printTime(times):
    # Clear previous outputs
    clear()

    for t in times:
        print(t[0], convertTime(t[1]))


def convertTime(time):
    return str(round(time * 1000, 2)) + "ms"


def updateTime(timePrev, timeBefore, timeAfter):
    timeNew = abs(timeAfter - timeBefore)
    return (timePrev + timeNew) / 2


cap = cv.VideoCapture(VIDEO_NAME)
cap.set(cv.CAP_PROP_POS_FRAMES, FRAME_START_INDEX)

matchesBetweenFrames = []


def computeSpeedAndDistance(running, frameIndex, speedX, speedY, x, y):
    # Local variables
    speedPrevX, speedPrevY = 0, 0
    kp, des = None, None
    kpTime = 0
    matchTime = 0
    previousIndex = 0

    # It's own capture reading from video to avoid race conditions in reading
    cap = cv.VideoCapture(VIDEO_NAME)
    cap.set(cv.CAP_PROP_POS_FRAMES, FRAME_START_INDEX)

    # Initial frame reading
    _, framePrev = cap.read()
    # Initial previous keypoints and descriptors
    kpPrev, desPrev = findKeyPoints(framePrev, KEYPOINTS_METHOD)

    while running.value == 1:
        frameIndexCurrent = frameIndex.value
        cap.set(cv.CAP_PROP_POS_FRAMES, frameIndexCurrent + FRAME_START_INDEX)

        numOfFramesBetweenEstimation = frameIndexCurrent - previousIndex

        # print("Calculating with index", frameIndex.value)

        _, frame = cap.read()

        timeBeforeKp = time.time()
        kp, des = findKeyPoints(frame, KEYPOINTS_METHOD)
        timeAfterKp = time.time()

        timeBeforeMatching = time.time()
        matchesBetweenFrames = findMatches(frame, framePrev, kp, kpPrev, des, desPrev)
        timeAfterMatching = time.time()

        distanceX, distanceY = distanceBetweenMatches(matchesBetweenFrames, kp, kpPrev)

        newSpeedX, speedPrevX = calculateSpeed(
            distanceX, speedPrevX, numOfFramesBetweenEstimation
        )

        newSpeedY, speedPrevY = calculateSpeed(
            distanceY, speedPrevY, numOfFramesBetweenEstimation
        )

        newX, newY = calculatePosition(
            x.value, y.value, newSpeedX, newSpeedY, numOfFramesBetweenEstimation
        )

        # Update global shared variable
        speedX.value = newSpeedX
        speedY.value = newSpeedY

        x.value = newX
        y.value = newY

        # x, y = calculatePosition(x, y, newSpeedX, newSpeedY, NUM_OF_FRAMES_BETWEEN_ESTIMATION)

        kpTime = updateTime(kpTime, timeBeforeKp, timeAfterKp)
        matchTime = updateTime(matchTime, timeBeforeMatching, timeAfterMatching)
        print(printTime([["Keypoints time:", kpTime], ["Matching time:", matchTime]]))

        frameCopy = frame.copy()
        framePrevCopy = framePrev.copy()

        # Add text to show which frame is which
        cv.putText(
            frameCopy,
            "Current frame",
            (50, 50),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )
        cv.putText(
            frameCopy,
            "Frame index: " + str(frameIndexCurrent),
            (50, 90),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )

        cv.putText(
            framePrevCopy,
            "Frame index: " + str(previousIndex),
            (50, 90),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )

        cv.putText(
            framePrevCopy,
            "Previous frame",
            (50, 50),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )

        # Need to draw only good matches, so create a mask
        matchesMask = [[0, 0] for i in range(len(matchesBetweenFrames))]

        # ratio test as per Lowe's paper
        for i, (m, n) in enumerate(matchesBetweenFrames):
            if m.distance < 0.7 * n.distance:
                matchesMask[i] = [1, 0]

        draw_params = dict(
            matchColor=(0, 255, 0),
            singlePointColor=(0, 0, 255),
            matchesMask=matchesMask,
            flags=0,
        )

        matchesFrame = cv.drawMatchesKnn(
            framePrevCopy,
            kpPrev,
            frameCopy,
            kp,
            matchesBetweenFrames,
            None,
            **draw_params
        )

        cv.imshow("matches with drawing", rescaleFrame(matchesFrame, 0.4))

        k = cv.waitKey(30) & 0xFF

        # Show the Output
        if k == 27:
            running.value = 0
            break

        framePrev = frame
        kpPrev, desPrev = kp, des
        previousIndex = frameIndexCurrent


def drawFrames(running, frameIndex, speedX, speedY, x, y):
    _, framePrev = cap.read()
    kpPrev, desPrev = findKeyPoints(framePrev, KEYPOINTS_METHOD)

    kpTime = 0
    matchTime = 0
    drawingTime = 0
    totalPerFrame = 0

    while True:
        # print("Frame running!")
        # if frameIndex == 20:
        #     running.value = 0
        #     break

        _, frame = cap.read()

        matchesFrame = frame

        # if not running:
        #     # print("new thread")
        #     t2 = Thread(None, computeSpeedAndDistance, None, [])
        #     running = True
        #     t2.start()

        # else:
        # Parallell testing
        frameCopy = frame.copy()
        framePrevCopy = framePrev.copy()

        # ---------------------------------------
        # -------- DRAWING AND RENDERING --------
        # ---------------------------------------

        # Add text to show which frame is which
        cv.putText(
            frameCopy,
            "Current frame",
            (50, 50),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )
        cv.putText(
            frameCopy,
            "Frame index: " + str(frameIndex.value),
            (50, 90),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )

        cv.putText(
            framePrevCopy,
            "Previous frame",
            (50, 50),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )

        # Add speed and positon text
        cv.putText(
            frameCopy,
            "SpeedX: {0:.2g}".format(speedX.value) + "m/s",
            (210, 180),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (0, 255, 255),
            2,
        )

        cv.putText(
            frameCopy,
            "SpeedY: {0:.2g}".format(speedY.value) + "m/s",
            (210, 225),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (0, 255, 255),
            2,
        )

        cv.putText(
            frameCopy,
            "Position: X{0:.2g}".format(x.value) + "m, Y{0:.2g}".format(y.value) + "m",
            (210, 265),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (0, 255, 255),
            2,
        )

        timeBeforeDrawing = time.time()
        # Need to draw only good matches, so create a mask
        matchesMask = [[0, 0] for i in range(len(matchesBetweenFrames))]

        # ratio test as per Lowe's paper
        for i, (m, n) in enumerate(matchesBetweenFrames):
            if m.distance < 0.7 * n.distance:
                matchesMask[i] = [1, 0]

        draw_params = dict(
            matchColor=(0, 255, 0),
            singlePointColor=(0, 0, 255),
            matchesMask=matchesMask,
            flags=0,
        )

        # matchesFrame = cv.drawMatchesKnn(
        #     framePrevCopy,
        #     kpPrev,
        #     frameCopy,
        #     kp,
        #     matchesBetweenFrames,
        #     None,
        #     **draw_params
        # )
        timeAfterDrawing = time.time()

        # kpTime = updateTime(kpTime, timeBeforeKp, timeAfterKp)
        # matchTime = updateTime(matchTime, timeBeforeMatching, timeAfterMatching)
        drawingTime = updateTime(drawingTime, timeBeforeDrawing, timeAfterDrawing)
        totalPerFrame = kpTime + matchTime + drawingTime

        times = [
            ["Keypoints time:", kpTime],
            ["Matching time:", matchTime],
            ["Drawing time", drawingTime],
            ["Total per frame", totalPerFrame],
        ]

        # printTime(times)

        # else:
        #     distanceBetweenEstimation += distance

        frameIndex.value = frameIndex.value + 1

        cv.imshow("matches", rescaleFrame(frameCopy, 0.5))
        # cv.imshow("frame", rescaleFrame(frame, 0.4))

        k = cv.waitKey(30) & 0xFF

        # Show the Output
        if k == 27:
            running.value = 0
            break


if __name__ == "__main__":
    # Shared variable between processes
    # If True (1), all processes will terminate
    running = mp.Value("i", 1)
    frameIndex = mp.Value("i", 0)
    speedX = mp.Value("d", 0)
    speedY = mp.Value("d", 0)
    x = mp.Value("d", 0)
    y = mp.Value("d", 0)

    processes = []
    tasks = [drawFrames, computeSpeedAndDistance]

    for t in tasks:
        p = mp.Process(target=t, args=(running, frameIndex, speedX, speedY, x, y))
        processes.append(p)
        p.start()

    for p in processes:
        p.join()
