import inspect
import math
import pprint
import statistics
import time
import sys
import csv
import cv2 as cv

# Fix path when importing files from different folder
sys.path.append("./")
from feature import (
    findKeyPoints,
    findMatches,
    filterMatches,
    distanceBetweenMatches,
)
from utils import rescaleFrame


def writeDataToFile(data):
    with open("./Testing/Results/test1.csv", "w", newline="") as file:
        # Open CSV writer with delimiter (value seperator) as ;
        writer = csv.writer(file, delimiter=";")

        for key, value in data.items():
            writer.writerow([key, str(format(value, ".2f")).replace(".", ",")])


# Constants
VIDEO_NAME = "sim_realtime.mp4"
VIDEO_FPS = 30
VIDEO_START_INDEX = 0
NUM_OF_FRAMES_TO_SKIP = 5
KEYPOINTS_METHOD = "ORB"
REFERENCE_SPEED = 0.2  # In meters per second

# Find total number of frames in video
cap = cv.VideoCapture("./Resources/" + VIDEO_NAME)
NUM_OF_FRAMES_IN_VIDEO = int(cap.get(cv.CAP_PROP_FRAME_COUNT))

# 3cm distance between squares in net
DISTANCE_BETWEEN_SQUARES = 0.03
NUM_OF_SQUARES_IN_FRAME = 15.5
VIDEO_HEIGHT = 720
DISTANCE_PER_PX = (DISTANCE_BETWEEN_SQUARES * NUM_OF_SQUARES_IN_FRAME) / VIDEO_HEIGHT

REFERENCE_DISTANCE = (5 / 30) * REFERENCE_SPEED

# Open capture of video at the set start index
cap = cv.VideoCapture("./Resources/" + VIDEO_NAME)
cap.set(cv.CAP_PROP_POS_FRAMES, VIDEO_START_INDEX)

# Read first frame from capture
_, frame1 = cap.read()

# Skip to next frame
cap.set(cv.CAP_PROP_POS_FRAMES, VIDEO_START_INDEX + NUM_OF_FRAMES_TO_SKIP)

# Read second frame from capture
_, frame2 = cap.read()

# frame1 = frame1[0:200, 100:400]
# frame2 = frame2[0:200, 100:400]

# Array with all distances between matches
distances = []
matchesArr = []
matchesUpwards = []
matchesDownwards = []

for i in range(0, 1):
    # Find keypoints and descriptors for the first frame
    kp1, des1 = findKeyPoints(frame1, KEYPOINTS_METHOD)

    # Find keypoints and descriptors for the second frame
    kp2, des2 = findKeyPoints(frame2, KEYPOINTS_METHOD)

    # Find matches between frames
    matches = findMatches(frame1, frame2, kp1, kp2, des1, des2)

    # -- Calculate distance between matches --
    for m in matches:
        # Get indexes for keypoints from match object
        index1 = m[0].trainIdx
        index2 = m[0].queryIdx

        # Calculate distances between keypoints based on index
        distance = kp1[index1].pt[1] - kp2[index2].pt[1]

        # Add distance to data array
        distances.append(distance * DISTANCE_PER_PX)

        matchesArr.append(m)

        if distance > 0:
            matchesUpwards.append(m)
        else:
            matchesDownwards.append(m)


# Find the average distance
totalDistance = 0
for d in distances:
    totalDistance += abs(d)
distanceAvg = totalDistance / len(distances)

# Other Distance statistical calculations
distanceMin = min(distances)
distanceMax = max(distances)
distanceVar = statistics.variance(distances)
distanceDev = statistics.pstdev(distances)
distanceRelDev = (100 * distanceDev) / distanceAvg
precision = distanceDev / (distanceMax - distanceMin)

# Calculate relative error with reference value and average
relError = (abs(distanceAvg - REFERENCE_DISTANCE)) / abs(REFERENCE_DISTANCE)


# Calculate accuracy with reference value and average
accuracy = (
    REFERENCE_DISTANCE
    - (max(REFERENCE_DISTANCE, distanceAvg) - min(REFERENCE_DISTANCE, distanceAvg))
) / REFERENCE_DISTANCE

# If the calculated accuracy is outside of the 0-100% range of the reference value
# the accuracy must be 0
if accuracy < 0:
    accuracy = 0

# Convert distance data to a more readable format
distanceAvg = round(distanceAvg, 2)
distanceMin = round(distanceMin, 2)
distanceMax = round(distanceMax, 2)
distanceVar = round(distanceVar, 2)
distanceDev = round(distanceDev, 2)
referenceDistance = round(REFERENCE_DISTANCE, 2)
accuracy = round(accuracy * 100, 2)
precision = round(precision * 100, 2)
relError = round(relError * 100, 2)

# Save all values in dictionary
data = {
    "Avstand referanse (m)": referenceDistance,
    "Avstand gjennomsnitt (m)": distanceAvg,
    "Avstand min (m)": distanceMin,
    "Avstand max (m)": distanceMax,
    "Avstand varianse (m)": distanceVar,
    "Avstand standardavvik (m)": distanceDev,
    "Presisjon (%)": precision,
    "Nøyaktighet (%)": accuracy,
    "Relativ feil (%)": relError,
    "Antall matcher nedover": len(matchesDownwards),
    "Antall matcher oppover": len(matchesUpwards),
}

draw_params = dict(
    flags=2,
)

matchesFrame = cv.drawMatchesKnn(
    frame1, kp1, frame2, kp2, matchesArr, None, **draw_params
)
cv.imshow("matches", rescaleFrame(matchesFrame, 0.6))
cv.waitKey(0)

writeDataToFile(data)